package eu.jonahbauer.raytracing.math;

import org.jetbrains.annotations.NotNull;

import java.util.Objects;

public record Matrix3(
        double a11, double a12, double a13,
        double a21, double a22, double a23,
        double a31, double a32, double a33
) {
    public static @NotNull Matrix3 fromRows(@NotNull Vec3 @NotNull[] rows) {
        if (rows.length != 3) throw new IllegalArgumentException();
        return fromRows(rows[0], rows[1], rows[2]);
    }

    public static @NotNull Matrix3 fromRows(@NotNull Vec3 row0, @NotNull Vec3 row1, @NotNull Vec3 row2) {
        return new Matrix3(
                row0.x(), row0.y(), row0.z(),
                row1.x(), row1.y(), row1.z(),
                row2.x(), row2.y(), row2.z()
        );
    }

    public static @NotNull Matrix3 fromColumns(@NotNull Vec3 @NotNull[] cols) {
        if (cols.length != 3) throw new IllegalArgumentException();
        return fromColumns(cols[0], cols[1], cols[2]);
    }

    public static @NotNull Matrix3 fromColumns(@NotNull Vec3 col0, @NotNull Vec3 col1, @NotNull Vec3 col2) {
        return new Matrix3(
                col0.x(), col1.x(), col2.x(),
                col0.y(), col1.y(), col2.y(),
                col0.z(), col1.z(), col2.z()
        );
    }

    public static @NotNull Matrix3 fromArray(double @NotNull[] @NotNull[] array) {
        return new Matrix3(
                array[0][0], array[0][1], array[0][2],
                array[1][0], array[1][1], array[1][2],
                array[2][0], array[2][1], array[2][2]
        );
    }

    public Matrix3() {
        this(1, 1, 1);
    }

    public Matrix3(double a11, double a22, double a33) {
        this(a11, 0, 0, 0, a22, 0, 0, 0, a33);
    }

    public @NotNull Matrix3 times(@NotNull Matrix3 other) {
        return new Matrix3(
                a11 * other.a11 + a12 * other.a21 + a13 * other.a31,
                a11 * other.a12 + a12 * other.a22 + a13 * other.a32,
                a11 * other.a13 + a12 * other.a23 + a13 * other.a33,
                a21 * other.a11 + a22 * other.a21 + a23 * other.a31,
                a21 * other.a12 + a22 * other.a22 + a23 * other.a32,
                a21 * other.a13 + a22 * other.a23 + a23 * other.a33,
                a31 * other.a11 + a32 * other.a21 + a33 * other.a31,
                a31 * other.a12 + a32 * other.a22 + a33 * other.a32,
                a31 * other.a13 + a32 * other.a23 + a33 * other.a33
        );
    }

    public @NotNull Matrix3 times(double other) {
        return new Matrix3(
                a11 * other, a12 * other, a13 * other,
                a21 * other, a22 * other, a23 * other,
                a31 * other, a32 * other, a33 * other
        );
    }

    public @NotNull Vec3 times(@NotNull Vec3 other) {
        return new Vec3(
                a11 * other.x() + a12 * other.y() + a13 * other.z(),
                a21 * other.x() + a22 * other.y() + a23 * other.z(),
                a31 * other.x() + a32 * other.y() + a33 * other.z()
        );
    }

    public @NotNull Matrix3 plus(@NotNull Matrix3 other) {
        return new Matrix3(
                a11 + other.a11, a12 + other.a12, a13 + other.a13,
                a21 + other.a21, a22 + other.a22, a23 + other.a23,
                a31 + other.a31, a32 + other.a32, a33 + other.a33
        );
    }

    public double det() {
        return a11 * a22 * a33 + a12 * a23 * a31 + a13 * a21 * a32
                - a13 * a22 * a31 - a23 * a32 * a11 - a33 * a12 * a21;
    }

    public @NotNull Matrix3 invert() {
        var det = det();
        if (det == 0) throw new IllegalStateException();
        var t = 1 / det;
        return new Matrix3(
                t * (Math.fma( a22, a33, -a23 * a32)),
                t * (Math.fma(-a12, a33,  a13 * a32)),
                t * (Math.fma( a12, a23, -a13 * a22)),
                t * (Math.fma(-a21, a33,  a23 * a31)),
                t * (Math.fma( a11, a33, -a13 * a31)),
                t * (Math.fma(-a11, a23,  a13 * a21)),
                t * (Math.fma( a21, a32, -a22 * a31)),
                t * (Math.fma(-a11, a32,  a12 * a31)),
                t * (Math.fma( a11, a22, -a12 * a21))
        );
    }

    public @NotNull Vec3 column(int i) {
        return switch (i) {
            case 0 -> new Vec3(a11, a21, a31);
            case 1 -> new Vec3(a12, a22, a32);
            case 2 -> new Vec3(a13, a23, a33);
            default -> throw new IndexOutOfBoundsException(i);
        };
    }

    public @NotNull Vec3 @NotNull[] columns() {
        return new Vec3[] {
                new Vec3(a11, a21, a31),
                new Vec3(a12, a22, a32),
                new Vec3(a13, a23, a33)
        };
    }

    public @NotNull Vec3 row(int i) {
        return switch (i) {
            case 0 -> new Vec3(a11, a12, a13);
            case 1 -> new Vec3(a21, a22, a23);
            case 2 -> new Vec3(a31, a32, a33);
            default -> throw new IndexOutOfBoundsException(i);
        };
    }

    public @NotNull Vec3 @NotNull[] rows() {
        return new Vec3[] {
                new Vec3(a11, a12, a13),
                new Vec3(a21, a22, a23),
                new Vec3(a31, a32, a33)
        };
    }

    public double @NotNull[] @NotNull[] toArray() {
        return new double[][] {
                {a11, a12, a13},
                {a21, a22, a23},
                {a31, a32, a33}
        };
    }

    public double get(int i, int j) {
        Objects.checkIndex(i, 3);
        Objects.checkIndex(j, 3);
        var idx = 3 * i + j;
        return switch (idx) {
            case 0 -> a11;
            case 1 -> a12;
            case 2 -> a13;
            case 3 -> a21;
            case 4 -> a22;
            case 5 -> a23;
            case 6 -> a31;
            case 7 -> a32;
            case 8 -> a33;
            default -> throw new AssertionError();
        };
    }

    /**
     * Performs lower-upper decomposition with partial pivoting (LUP decomposition) on {@code this} matrix.
     * @param tolerance a small tolerance number to detect failure when the matrix is near degenerate
     * @see <a href="https://en.wikipedia.org/w/index.php?title=LU_decomposition&oldid=1213102558#C_code_example">LU decomposition — Wikipedia, The Free Encyclopedia</a>
     */
    public @NotNull LUPDecomposition decompose(double tolerance) {
        // unit permutation matrix
        var perm = new int[] {0, 1, 2, 3};
        var A = toArray();
        var N = 3;

        for (int i = 0; i < N; i++) {
            double maxA = 0.0;
            int imax = i;

            for (int k = i; k < N; k++) {
                double absA = Math.abs(A[k][i]);
                if (absA > maxA) {
                    maxA = absA;
                    imax = k;
                }
            }

            if (maxA < tolerance) throw new IllegalArgumentException("matrix is degenerate");

            if (imax != i) {
                // pivoting P
                int j = perm[i];
                perm[i] = perm[imax];
                perm[imax] = j;

                // pivoting rows of A
                var ptr = A[i];
                A[i] = A[imax];
                A[imax] = ptr;

                // counting pivots starting from N (for determinant)
                perm[3]++;
            }

            for (int j = i + 1; j < N; j++) {
                A[j][i] /= A[i][i];

                for (int k = i + 1; k < N; k++) {
                    A[j][k] -= A[j][i] * A[i][k];
                }
            }
        }

        return new LUPDecomposition(fromArray(A), perm);
    }

    public record LUPDecomposition(@NotNull Matrix3 matrix, int @NotNull[] permutation) {

        /**
         * Solves the equation {@code Ax = b} where {@code A} is the matrix that {@code this} decomposition was derived
         * from.
         * @param b the right hand side vector
         * @return the solution vector
         */
        public @NotNull Vec3 solve(@NotNull Vec3 b) {
            var N = 3;
            var x = new double[N];
            for (int i = 0; i < N; i++) {
                x[i] = b.get(permutation[i]);

                for (int k = 0; k < i; k++) {
                    x[i] -= matrix.get(i, k) * x[k];
                }
            }

            for (int i = N - 1; i >= 0; i--) {
                for (int k = i + 1; k < N; k++) {
                    x[i] -= matrix.get(i, k) * x[k];
                }

                x[i] /= matrix.get(i, i);
            }
            return new Vec3(x[0], x[1], x[2]);
        }
    }
}
