package eu.jonahbauer.raytracing.render.canvas;

import eu.jonahbauer.raytracing.render.spectrum.SampledSpectrum;
import eu.jonahbauer.raytracing.render.spectrum.SampledWavelengths;
import eu.jonahbauer.raytracing.render.color.ColorRGB;
import eu.jonahbauer.raytracing.render.color.ColorSpace;
import eu.jonahbauer.raytracing.render.color.ColorXYZ;
import org.jetbrains.annotations.NotNull;

import java.awt.image.BufferedImage;
import java.util.Objects;

public final class XYZCanvas implements Canvas {
    private final int width;
    private final int height;

    private final @NotNull ColorXYZ[][] data;

    public XYZCanvas(int width, int height) {
        this.width = width;
        this.height = height;

        if (width <= 0) throw new IllegalArgumentException("width must be positive");
        if (height <= 0) throw new IllegalArgumentException("height must be positive");

        this.data = new ColorXYZ[height][width];
    }

    public XYZCanvas(@NotNull BufferedImage image, @NotNull ColorSpace cs) {
        this(image.getWidth(), image.getHeight());

        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                data[y][x] = cs.toXYZ(new ColorRGB(image.getRGB(x, y)));
            }
        }
    }

    @Override
    public int getWidth() {
        return width;
    }

    @Override
    public int getHeight() {
        return height;
    }

    @Override
    public void add(int x, int y, int n, @NotNull SampledSpectrum spectrum, @NotNull SampledWavelengths lambda) {
        assert x < width;
        assert y < height;

        var xyz = spectrum.toXYZ(lambda);
        data[y][x] = ColorXYZ.average(get(x, y), xyz, n);
    }

    public @NotNull ColorXYZ get(int x, int y) {
        assert x < width;
        assert y < height;
        return Objects.requireNonNullElse(data[y][x], ColorXYZ.BLACK);
    }

    @Override
    public @NotNull ColorRGB getRGB(int x, int y, @NotNull ColorSpace cs) {
        return cs.toRGB(get(x, y));
    }

    public void set(int x, int y, @NotNull ColorXYZ color) {
        assert x < width;
        assert y < height;
        data[y][x] = Objects.requireNonNull(color);
    }
}
