package eu.jonahbauer.raytracing.render.color;

import eu.jonahbauer.raytracing.math.IVec3;
import org.jetbrains.annotations.NotNull;

import java.util.Random;

public record ColorRGB(double r, double g, double b) implements IVec3<ColorRGB> {
    public static final @NotNull ColorRGB BLACK = new ColorRGB(0.0, 0.0, 0.0);
    public static final @NotNull ColorRGB WHITE = new ColorRGB(1.0, 1.0, 1.0);

    public static @NotNull ColorRGB random(@NotNull Random random) {
        return new ColorRGB(random.nextDouble(), random.nextDouble(), random.nextDouble());
    }

    public static @NotNull ColorRGB random(@NotNull Random random, double min, double max) {
        var span = max - min;
        return new ColorRGB(
                Math.fma(random.nextDouble(), span, min),
                Math.fma(random.nextDouble(), span, min),
                Math.fma(random.nextDouble(), span, min)
        );
    }

    public ColorRGB(int rgb) {
        this((rgb >> 16) & 0xFF, (rgb >> 8) & 0xFF, rgb & 0xFF);
    }
    
    public ColorRGB(int red, int green, int blue) {
        this(red / 255f, green / 255f, blue / 255f);
    }

    public ColorRGB {
        assert Double.isFinite(r) : "r must be finite";
        assert Double.isFinite(g) : "g must be finite";
        assert Double.isFinite(b) : "b must be finite";
    }

    /*
     * Math
     */

    public static @NotNull ColorRGB average(@NotNull ColorRGB current, @NotNull ColorRGB next, int index) {
        return lerp(current, next, 1d / index);
    }

    public static @NotNull ColorRGB lerp(@NotNull ColorRGB a, @NotNull ColorRGB b, double t) {
        if (t < 0) return a;
        if (t > 1) return b;
        return new ColorRGB(
                Math.fma(t, b.r - a.r, a.r),
                Math.fma(t, b.g - a.g, a.g),
                Math.fma(t, b.b - a.b, a.b)
        );
    }

    public static @NotNull ColorRGB fma(@NotNull ColorRGB a, @NotNull ColorRGB b, @NotNull ColorRGB c) {
        return new ColorRGB(
                Math.fma(a.r, b.r, c.r),
                Math.fma(a.g, b.g, c.g),
                Math.fma(a.b, b.b, c.b)
        );
    }

    public static @NotNull ColorRGB gamma(@NotNull ColorRGB color) {
        return new ColorRGB(gamma(color.r), gamma(color.g), gamma(color.b));
    }

    public static @NotNull ColorRGB inverseGamma(@NotNull ColorRGB color) {
        return new ColorRGB(inverseGamma(color.r), inverseGamma(color.g), inverseGamma(color.b));
    }

    private static double gamma(double value) {
        if (value <= 0.0031308) return 12.92 * value;
        return 1.055 * Math.pow(value, 1. / 2.4) - 0.055;
    }

    private static double inverseGamma(double value) {
        if (value <= 0.04045) return value / 12.92;
        return Math.pow((value + 0.055) / 1.055, 2.4d);
    }

    @Override
    public @NotNull ColorRGB plus(@NotNull ColorRGB other) {
        return new ColorRGB(r + other.r, g + other.g, b + other.b);
    }

    @Override
    public @NotNull ColorRGB minus(@NotNull ColorRGB other) {
        return new ColorRGB(r - other.r, g - other.g, b - other.b);
    }

    @Override
    public @NotNull ColorRGB times(double d) {
        return new ColorRGB(r * d, g * d, b * d);
    }

    @Override
    public @NotNull ColorRGB times(@NotNull ColorRGB other) {
        return new ColorRGB(r * other.r, g * other.g, b * other.b);
    }

    /*
     * Accessors
     */

    public int red() {
        return toInt(r);
    }

    public int green() {
        return toInt(g);
    }

    public int blue() {
        return toInt(b);
    }

    private static int toInt(double value) {
        return Math.clamp((int) (255.99 * value), 0, 255);
    }

    @Override
    public double component1() {
        return r;
    }

    @Override
    public double component2() {
        return g;
    }

    @Override
    public double component3() {
        return b;
    }
}
