package eu.jonahbauer.raytracing.render.spectrum;

import eu.jonahbauer.raytracing.math.IVec;
import eu.jonahbauer.raytracing.render.color.ColorSpace;
import eu.jonahbauer.raytracing.render.color.ColorXYZ;
import eu.jonahbauer.raytracing.render.color.ColorRGB;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;

// TODO use Vector API to parallelize operations
public final class SampledSpectrum implements IVec<SampledSpectrum> {
    public static final SampledSpectrum BLACK;
    public static final SampledSpectrum WHITE;

    static {
        BLACK = new SampledSpectrum(new double[SampledWavelengths.SAMPLES]);
        var one = new double[SampledWavelengths.SAMPLES];
        Arrays.fill(one, 1);
        WHITE = new SampledSpectrum(one);
    }

    private final double @NotNull[] values;

    public SampledSpectrum(@NotNull SampledWavelengths lambdas, @NotNull Spectrum spectrum) {
        var values = new double[lambdas.size()];
        for (int i = 0; i < values.length; i++) {
            values[i] = spectrum.get(lambdas.get(i));
        }
        this.values = values;
    }

    private SampledSpectrum(double @NotNull[] values) {
        this.values = values;
    }

    /*
     * Math
     */

    public static @NotNull SampledSpectrum fma(@NotNull SampledSpectrum a, @NotNull SampledSpectrum b, @NotNull SampledSpectrum c) {
        var out = new double[a.values.length];
        for (int i = 0; i < a.values.length; i++) {
            out[i] = Math.fma(a.values[i], b.values[i], c.values[i]);
        }
        return new SampledSpectrum(out);
    }

    public static @NotNull SampledSpectrum lerp(@NotNull SampledSpectrum a, @NotNull SampledSpectrum b, double t) {
        if (t < 0) return a;
        if (t > 1) return b;
        var out = new double[a.values.length];
        for (int i = 0; i < a.values.length; i++) {
            out[i] = Math.fma(t, b.values[i] - a.values[i], a.values[i]);
        }
        return new SampledSpectrum(out);
    }

    @Override
    public @NotNull SampledSpectrum plus(@NotNull SampledSpectrum other) {
        var out = new double[other.values.length];
        for (int i = 0; i < other.values.length; i++) {
            out[i] = values[i] + other.values[i];
        }
        return new SampledSpectrum(out);
    }

    @Override
    public @NotNull SampledSpectrum minus(@NotNull SampledSpectrum other) {
        var out = new double[other.values.length];
        for (int i = 0; i < other.values.length; i++) {
            out[i] = values[i] - other.values[i];
        }
        return new SampledSpectrum(out);
    }

    @Override
    public @NotNull SampledSpectrum times(@NotNull SampledSpectrum other) {
        var out = new double[other.values.length];
        for (int i = 0; i < other.values.length; i++) {
            out[i] = values[i] * other.values[i];
        }
        return new SampledSpectrum(out);
    }

    @Override
    public @NotNull SampledSpectrum times(double d) {
        var out = new double[values.length];
        for (int i = 0; i < values.length; i++) {
            out[i] = values[i] * d;
        }
        return new SampledSpectrum(out);
    }

    @Override
    public double @NotNull [] toArray() {
        return Arrays.copyOf(values, values.length);
    }

    /*
     * Accessors
     */

    @Override
    public double get(int index) {
        return values[index];
    }

    public int size() {
        return values.length;
    }

    /*
     * Object
     */

    @Override
    public boolean equals(Object obj) {
        return obj instanceof SampledSpectrum o && Arrays.equals(values, o.values);
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(values);
    }

    @Override
    public @NotNull String toString() {
        return "SampledSpectrum[values=" + Arrays.toString(values) + "]";
    }

    /*
     * Conversions
     */

    public @NotNull ColorXYZ toXYZ(@NotNull SampledWavelengths lambdas) {
        return lambdas.toXYZ(this);
    }

    public @NotNull ColorRGB toRGB(@NotNull SampledWavelengths lambdas, @NotNull ColorSpace cs) {
        return cs.toRGB(toXYZ(lambdas));
    }
}
