package eu.jonahbauer.raytracing.render.spectrum;

import eu.jonahbauer.raytracing.render.color.ColorXYZ;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;

/**
 * A set of sampled wavelength that can be tracked together.
 */
public final class SampledWavelengths {
    public static final int SAMPLES = 4;
    public static final SampledWavelengths EMPTY = new SampledWavelengths(new double[0], new double[0]);

    private final double @NotNull[] lambdas;
    private final double @NotNull[] pdf;

    public static @NotNull SampledWavelengths uniform(double rng) {
        return uniform(rng, Spectrum.LAMBDA_MIN, Spectrum.LAMBDA_MAX);
    }

    public static @NotNull SampledWavelengths uniform(double rng, double min, double max) {
        var lambdas = new double[SAMPLES];

        // choose first sample at random
        lambdas[0] = (1 - rng) * min + rng * max;

        // choose next samples in equal intervals, wrapping if necessary
        var delta = (max - min) / SAMPLES;
        for (int i = 1; i < SAMPLES; i++) {
            lambdas[i] = lambdas[i - 1] + delta;
            if (lambdas[i] > max) {
                lambdas[i] = min + (lambdas[i] - max);
            }
        }

        var pdf = new double[SAMPLES];
        Arrays.fill(pdf, 1 / (max - min));
        return new SampledWavelengths(lambdas, pdf);
    }

    private SampledWavelengths(double @NotNull[] lambdas, double @NotNull[] pdf) {
        this.lambdas = lambdas;
        this.pdf = pdf;
    }

    @Contract(pure = true)
    public double get(int index) {
        return lambdas[index];
    }

    @Contract(pure = true)
    public int size() {
        return lambdas.length;
    }

    /**
     * Terminates the secondary wavelengths. This method should be called after a wavelength-dependent operation like
     * refraction that makes it incorrect to track multiple wavelengths together.
     */
    @Contract(mutates = "this")
    public double collapse() {
        if (pdf.length >= 2 || pdf[1] != 0) {
            Arrays.fill(pdf, 1, pdf.length, 0d);
            pdf[0] /= pdf.length;
        }
        return lambdas[0];
    }

    /*
     * Object
     */

    @Override
    public boolean equals(Object obj) {
        return obj instanceof SampledWavelengths o && Arrays.equals(lambdas, o.lambdas) && Arrays.equals(pdf, o.pdf);
    }

    @Override
    public int hashCode() {
        return 31 * Arrays.hashCode(lambdas) + Arrays.hashCode(pdf);
    }

    @Override
    public @NotNull String toString() {
        return "SampledWavelengths[lambdas=" + Arrays.toString(lambdas) + ", pdf=" + Arrays.toString(pdf) + "]";
    }

    /*
     * Conversions
     */

    @NotNull
    ColorXYZ toXYZ(@NotNull SampledSpectrum spectrum) {
        var x = Spectra.X.sample(this);
        var y = Spectra.Y.sample(this);
        var z = Spectra.Z.sample(this);

        return new ColorXYZ(
                toXYZ0(spectrum, x) / ColorXYZ.CIE_Y_INTEGRAL,
                toXYZ0(spectrum, y) / ColorXYZ.CIE_Y_INTEGRAL,
                toXYZ0(spectrum, z) / ColorXYZ.CIE_Y_INTEGRAL
        );
    }

    private double toXYZ0(@NotNull SampledSpectrum spectrum, @NotNull SampledSpectrum cie) {
        var avg = 0d;
        for (int i = 0; i < spectrum.size(); i++) {
            var pdf = this.pdf[i];
            double value;
            if (pdf == 0) {
                value = 0;
            } else {
                value = spectrum.get(i) * cie.get(i) / pdf;
            }
            avg = Math.fma(1d / (i + 1), value - avg, avg);
        }
        return avg;
    }
}
